// asset.hpp
//
// Copyright (C) 2006 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#ifndef ASSET_INC_
#define ASSET_INC_

#ifdef _MSC_VER
#pragma warning ( push )
#pragma warning ( disable: 4701 4127 4511 4512 4251 )
#endif

#include <map>

#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>

#include <openpluginlib/pl/openpluginlib.hpp>
#include <openpluginlib/pl/utf8_utils.hpp>

#include <openassetlib/al/config.hpp>

#if defined ( HAVE_FLEX_STRING )

// flex_string doesn't implement operator>>. The below
// is a hack implementation for use with boost::lexical_cast.

template <typename E, class T, class A, class S>
std::basic_istream<typename flex_string<E, T, A, S>::value_type, typename flex_string<E, T, A, S>::traits_type>& 
operator>>(
    std::basic_istream<typename flex_string<E, T, A, S>::value_type,
    typename flex_string<E, T, A, S>::traits_type>& is,
    flex_string<E, T, A, S>& str )
{
	std::basic_string<E, T, A> tmp;
	
	is >> tmp;
	str.assign( tmp.begin( ), tmp.end( ) );
	
	return is;
}

#endif

namespace opl = olib::openpluginlib;

namespace olib { namespace openassetlib { namespace al {

typedef boost::shared_ptr<class asset> asset_ptr;

class AL_DECLSPEC asset
{
public:
	typedef std::map<opl::string, opl::string>	map_type;
	typedef map_type::allocator_type			allocator_type;
	typedef map_type::const_iterator			const_iterator;
	typedef map_type::const_reference			const_reference;
	typedef map_type::const_reverse_iterator	const_reverse_iterator;
	typedef map_type::difference_type			difference_type;
	typedef map_type::iterator					iterator;
	typedef map_type::key_compare				key_compare;
	typedef map_type::key_type					key_type;
	typedef map_type::mapped_type				mapped_type;
	typedef map_type::pointer					pointer;
	typedef map_type::reference					reference;
	typedef map_type::reverse_iterator			reverse_iterator;
	typedef map_type::size_type					size_type;
	typedef map_type::value_type				value_type;
	typedef map_type::value_compare				value_compare;

	static asset_ptr create()
	{
		return asset_ptr(new asset());
	}

	static asset_ptr create(const opl::string& guid)
	{
		return asset_ptr(new asset(guid));
	}

	static asset_ptr create(const asset_ptr& a)
	{
		return asset_ptr(new asset(a));
	}

	virtual ~asset( )
	{
	}

	const opl::string guid() const
	{
		return guid_;
	}

	void guid(const opl::string& guid)
	{
		guid_ = guid;
		dirty_ = true;
	}

	bool is_dirty() const
	{
		return dirty_;
	}

	void clear_dirty_flag()
	{
		dirty_ = false;
	}

	void set(const key_type& key, const bool value)
	{
		set(key, opl::string(value ? "true" : "false"));
	}

	void set(const key_type& key, const unsigned char value)
	{
		set(key, boost::lexical_cast<opl::string, const unsigned char>(value));
	}

	void set(const key_type& key, const char value)
	{
		set(key, boost::lexical_cast<opl::string>(value));
	}

	void set(const key_type& key, const unsigned short value)
	{
		set(key, boost::lexical_cast<opl::string>(value));
	}

	void set(const key_type& key, const short value)
	{
		set(key, boost::lexical_cast<opl::string>(value));
	}

	void set(const key_type& key, const unsigned long value)
	{
		set(key, boost::lexical_cast<opl::string>(value));
	}

	void set(const key_type& key, const long value)
	{
		set(key, boost::lexical_cast<opl::string>(value));
	}

	void set(const key_type& key, const unsigned int value)
	{
		set(key, boost::lexical_cast<opl::string>(value));
	}

	void set(const key_type& key, const int value)
	{
		set(key, boost::lexical_cast<opl::string>(value));
	}

	void set(const key_type& key, const float value)
	{
		set(key, boost::lexical_cast<opl::string>(value));
	}

	void set(const key_type& key, const double value)
	{
		set(key, boost::lexical_cast<opl::string>(value));
	}

	void set(const key_type& key, const long double value)
	{
		set(key, boost::lexical_cast<opl::string>(value));
	}

	void set(const key_type& key, const long long value)
	{
		set(key, boost::lexical_cast<opl::string>(value));
	}

	void set(const key_type& key, const opl::string value)
	{
		if(!map_.insert(value_type(key, value)).second)
		{
			iterator pair = map_.find(key);
			pair->second = value;
		}

		dirty_ = true;
	}

	const mapped_type get(const key_type& key) const
	{
		const_iterator pair = map_.find(key);
		return (pair != map_.end() ? pair->second : mapped_type());
	}

	const_iterator								begin() const										{	return map_.begin();			}
	size_type									count(const key_type& key) const					{	return map_.count(key);			}
	bool										empty() const										{	return map_.empty();			}
	const_iterator								end() const											{	return map_.end();				}
	std::pair<const_iterator, const_iterator>	equal_range(const key_type& key) const				{	return map_.equal_range(key);	}
	const_iterator								find(const key_type& key) const						{	return map_.find(key);			}
	allocator_type								get_allocator() const								{	return map_.get_allocator();	}
	key_compare									key_comp() const									{	return map_.key_comp();			}
	const_iterator								lower_bound(const key_type& key) const				{	return map_.lower_bound(key);	}
	size_type									max_size() const									{	return map_.max_size();			}
	const_reverse_iterator						rbegin() const										{	return map_.rbegin();			}
	const_reverse_iterator						rend() const										{	return map_.rend();				}
	size_type									size() const										{	return map_.size();				}
	const_iterator								upper_bound(const key_type& key) const				{	return map_.upper_bound(key);	}
	value_compare								value_comp() const									{	return map_.value_comp();		}

	// additional helper functions to be exported to python interface
	value_type find_metadata(const key_type& key) const	{	
		const_iterator it = map_.find(key);
		if ( it != end( ) )
			return *it;
		else
			return value_type( );
	}


private:
	asset();
	asset(const opl::string& guid);
	asset(const asset_ptr& asset);

	opl::string	guid_;
	map_type	map_;
	bool		dirty_;
};

} } }

#ifdef _MSC_VER
#pragma warning ( pop )
#endif

#endif
