
// ml - A media library representation.

// Copyright (C) 2005-2006 VM Inc.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#ifndef OPENMEDIALIB_INPUT_INC_
#define OPENMEDIALIB_INPUT_INC_

#include <boost/shared_ptr.hpp>

namespace olib { namespace openmedialib { namespace ml {

typedef enum
{
	process_image = 0x1,
	process_audio = 0x2,
	process_next_intra = 0x4
}
process_flags;

class ML_DECLSPEC input_type 
{
	public:
		// Constructor/destructor
		explicit input_type( ) : m_position( 0 ), m_process( process_image | process_audio )  { }
		virtual ~input_type( ) { }

		// Basic information
		virtual const openpluginlib::wstring get_uri( ) const = 0;
		virtual const openpluginlib::wstring get_mime_type( ) const = 0;
		virtual bool has_video( ) const = 0;
		virtual bool has_audio( ) const = 0;

		// Audio/Visual
		virtual void get_fps( int &num, int &den ) const = 0;
		virtual int get_frames( ) const = 0;
		virtual bool is_seekable( ) const = 0;

		// Visual
		virtual void get_sar( int &num, int &den ) const = 0;
		virtual int get_video_streams( ) const = 0;
		virtual int get_width( ) const = 0;
		virtual int get_height( ) const = 0;

		// Audio
		virtual int get_audio_streams( ) const = 0;

		// Set video and audio streams
		virtual bool set_video_stream( const int ) { return false; }
		virtual bool set_audio_stream( const int ) { return false; }

		// Default number of frames
		virtual double fps( ) const
		{
			int num, den;
			get_fps( num, den );
			return den != 0 ? double( num ) / double( den ) : 1;
		}

		virtual double get_sar( ) const
		{
			int num, den;
			get_sar( num, den );
			return den != 0 ? double( num ) / double( den ) : 1;
		}

		virtual double get_duration( )
		{
			return double( get_frames( ) ) / fps( );
		}

		// Provides a hint to the input implementation - allows the user to say 
		// 'I only want image or audio, or just fetch the nearest intra frame to here' in the next fetch
		// NB: usage is not enforced - up to the implementation to decide whether they do it or not
		void set_process_flags( int flags )
		{
			m_process = flags;
		}

		int get_process_flags( )
		{
			return m_process;
		}

		// Default seek functionality
		virtual void seek( const int position, const bool relative = false )
		{
			if ( relative )
				m_position = ( m_position + position );
			else
				m_position = position;

			if ( m_position < 0 ) 
				m_position = 0;
			else if ( m_position >= int( get_frames( ) ) ) 
				m_position = int( get_frames( ) - 1 );
		}

		virtual void seek_time( const double time, const bool relative = false )
		{
			int position = int( time * fps( ) );
			seek( position, relative );
		}

		virtual int get_position( ) const
		{
			return m_position;
		}

		// Virtual frame fetch method
		virtual frame_type_ptr fetch( ) = 0;

		// Allow the app to refresh a cached frame
		virtual void refresh_cache( frame_type_ptr ) { }


	private:
		int m_position;
		int m_process;
};

typedef boost::shared_ptr< input_type > input_type_ptr;

} } }

#endif
